# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""PeriFlow Checkpoint SDK."""

# pylint: disable=line-too-long, arguments-differ, too-many-arguments, too-many-statements, too-many-locals, redefined-builtin

from __future__ import annotations

import os
from pathlib import Path
from typing import List, Optional, cast
from uuid import UUID

import yaml

from periflow.client.checkpoint import CheckpointClient, CheckpointFormClient
from periflow.client.credential import CredentialClient
from periflow.client.group import GroupProjectCheckpointClient
from periflow.cloud.storage import build_storage_client
from periflow.enums import CheckpointCategory, CheckpointDataType, CredType, StorageType
from periflow.errors import (
    CheckpointConversionError,
    InvalidConfigError,
    NotFoundError,
    NotSupportedCheckpointError,
    PeriFlowInternalError,
    TokenizerNotFoundError,
)
from periflow.logging import logger
from periflow.schema.resource.v1.checkpoint import V1Checkpoint
from periflow.sdk.resource.base import ResourceAPI
from periflow.utils.fs import (
    FileSizeType,
    attach_storage_path_prefix,
    download_file,
    expand_paths,
    get_file_info,
    strip_storage_path_prefix,
)
from periflow.utils.maps import cred_type_map, cred_type_map_inv
from periflow.utils.validate import (
    validate_checkpoint_attributes,
    validate_cloud_storage_type,
    validate_enums,
    validate_storage_region,
)


class Checkpoint(ResourceAPI[V1Checkpoint, UUID]):
    """Checkpoint resource API."""

    @staticmethod
    def create(
        name: str,
        credential_id: UUID,
        cloud_storage: StorageType,
        region: str,
        storage_name: str,
        storage_path: Optional[str] = None,
        iteration: Optional[int] = None,
        attr_file_path: Optional[str] = None,
    ) -> V1Checkpoint:
        """Creates a checkpoint by linking the existing cloud storage (e.g., AWS S3, GCS, Azure Blob Storage) with PeriFlow.

        Args:
            name (str): The name of checkpoint to create.
            credential_id (UUID): Credential ID to access the cloud storage.
            cloud_storage (StorageType): Cloud storage type.
            region (str): Cloud region.
            storage_name (str): Storage name (e.g., AWS S3 bucket name).
            storage_path (Optional[str], optional): Path to the storage object (e.g., AWS S3 bucket key). Defaults to None.
            iteration (Optional[int], optional): The iteration of the checkpoint. Defaults to None.
            attr_file_path (Optional[str], optional): Path to the checkpoint attribute YAML file. Defaults to None.

        Returns:
            V1Checkpoint: Created checkpoint object.

        Raises:
            InvalidConfigError: Raised when checkpoint attribute file located at `attr_file_path` has invalid YAML format. Also raised when the credential with `credential_id` is not for the cloud provider of `cloud_storage`. Also raised when `region` is invalid.
            NotSupportedError: Raised when `cloud_storage` is not supported yet.
            InvalidAttributesError: Raised when the checkpoint attributes described in `attr_file_path` is in the invalid format.

        Examples:
            Basic usage:

            ```python
            import periflow as pf

            # Set up PeriFlow context.
            pf.init(
                api_key="YOUR_PERILFOW_API_KEY",
                project_name="my-project",
            )

            # Create a checkpoint by linking an existing S3 bucket.
            checkpoint = pf.Checkpoint.create(
                name="my-checkpoint",
                credential_id="YOUR_CREDENTIAL_ID",
                cloud_stroage="s3",
                region="us-east-1",
                storage_name="my-bucket",
                storage_path="path/to/ckpt",
                attr_file_path="path/to/attr.yaml",
            )
            ```

            :::info
            You need to create a credential to access the S3 storage in advance.
            :::

            :::note
            An example of attribute files is as follows.
            You have the flexibility to modify the value of each field according to your preferences,
            but it is mandatory to fill in every field.

            ```yaml
            # blenderbot
            model_type: blenderbot
            dtype: fp16
            head_size: 80
            num_heads: 32
            hidden_size: 2560
            ff_intermediate_size: 10240
            num_encoder_layers: 2
            num_decoder_layers: 24
            max_input_length: 128
            max_output_length: 128
            vocab_size: 8008
            eos_token: 2
            decoder_start_token: 1

            # bloom
            model_type: bloom
            dtype: fp16
            head_size: 128
            num_heads: 32
            num_layers: 30
            max_length: 2048
            vocab_size: 250880
            eos_token: 2

            # gpt
            model_type: gpt
            dtype: fp16
            head_size: 64
            num_heads: 25
            num_layers: 48
            max_length: 1024
            vocab_size: 50257
            eos_token: 50256

            # gpt-j
            model_type: gpt-j
            dtype: fp16
            head_size: 256
            rotary_dim: 64
            num_heads: 16
            num_layers: 28
            max_length: 2048
            vocab_size: 50400
            eos_token: 50256

            # gpt-neox
            model_type: gpt-neox
            dtype: fp16
            head_size: 128
            rotary_dim: 32
            num_heads: 40
            num_layers: 36
            max_length: 2048
            vocab_size: 50280
            eos_token: 0

            # llama
            model_type: llama
            dtype: fp16
            head_size: 128
            rotary_dim: 128
            num_heads: 32
            num_kv_heads: 32
            num_layers: 32
            ff_intermediate_size: 11008
            max_length: 2048
            vocab_size: 32000
            eos_token: 1

            # opt
            model_type: opt
            dtype: fp16
            head_size: 128
            num_heads: 32
            num_layers: 32
            max_length: 2048
            vocab_size: 50272
            eos_token: 2

            # t5
            model_type: t5
            dtype: fp16
            head_size: 128
            num_heads: 32
            hidden_size: 1024
            ff_intermediate_size: 16384
            num_encoder_layers: 24
            num_decoder_layers: 24
            max_input_length: 512
            max_output_length: 512
            num_pos_emb_buckets: 32
            max_pos_distance: 128
            vocab_size: 32100
            eos_token: 1
            decoder_start_token: 0

            # t5-v1_1
            model_type: t5-v1_1
            dtype: fp16
            head_size: 64
            num_heads: 32
            hidden_size: 2048
            ff_intermediate_size: 5120
            num_encoder_layers: 24
            num_decoder_layers: 24
            max_input_length: 512
            max_output_length: 512
            num_pos_emb_buckets: 32
            max_pos_distance: 128
            vocab_size: 32128
            eos_token: 1
            decoder_start_token: 0
            ```
            :::

        """
        cloud_storage = validate_enums(cloud_storage, StorageType)
        validate_cloud_storage_type(cloud_storage)
        validate_storage_region(vendor=cloud_storage, region=region)

        attr = {}
        if attr_file_path is not None:
            try:
                with open(attr_file_path, "r", encoding="utf-8") as attr_f:
                    attr = yaml.safe_load(attr_f)
            except yaml.YAMLError as exc:
                raise InvalidConfigError(
                    f"The attribute YAML file has invalid format: {str(exc)}"
                ) from exc

        if attr:
            validate_checkpoint_attributes(attr)

        credential_client = CredentialClient()
        credential = credential_client.get_credential(credential_id)
        if credential["type"] != cred_type_map[CredType(cloud_storage.value)]:
            raise InvalidConfigError(
                "Credential type and cloud vendor mismatch: "
                f"{cred_type_map_inv[credential['type']]} and {cloud_storage.value}."
            )

        storage_helper = build_storage_client(
            cloud_storage, credential_json=credential["value"]
        )
        if storage_path is not None:
            storage_path = storage_path.strip("/")
        files = storage_helper.list_storage_files(storage_name, storage_path)
        if storage_path is not None:
            storage_name = f"{storage_name}/{storage_path}"

        checkpoint_client = GroupProjectCheckpointClient()
        dist_config = {
            "pp_degree": 1,
            "dp_degree": 1,
            "mp_degree": 1,
            "dp_mode": "allreduce",
            "parallelism_order": ["pp", "dp", "mp"],
        }
        raw_ckpt = checkpoint_client.create_checkpoint(
            name=name,
            vendor=cloud_storage,
            region=region,
            credential_id=credential_id,
            iteration=iteration,
            storage_name=storage_name,
            files=files,
            dist_config=dist_config,
            attributes=attr,
        )
        ckpt = V1Checkpoint.model_validate(raw_ckpt)
        if ckpt.forms:
            for file in ckpt.forms[0].files:
                file.path = strip_storage_path_prefix(file.path)
        return ckpt

    @staticmethod
    def list(
        category: Optional[CheckpointCategory] = None,
        limit: int = 20,
        deleted: bool = False,
    ) -> List[V1Checkpoint]:
        """Lists checkpoints.

        Args:
            category (Optional[CheckpointCategory], optional): Filters by category. Defaults to None.
            limit (int, optional): The maximum number of retrieved results. Defaults to 20.
            deleted (bool, optional): Filters only the deleted checkpoints. Defaults to False.

        Returns:
            List[V1Checkpoint]: A list of retrieved checkpoints.

        Examples:
            To get the latest 100 checkpoints:

            ```python
            import periflow as pf

            # Set up PeriFlow context.
            pf.init(
                api_key="YOUR_PERILFOW_API_KEY",
                project_name="my-project",
            )

            checkpoints = pf.Checkpoint.list(limit=100)
            ```

            To get the deleted checkpoints created by users:

            ```python
            checkpoints = pf.Checkpoint.list(
                category=CheckpointCategory.USER_PROVIDED, deleted=True
            )
            ```

        """
        client = GroupProjectCheckpointClient()
        checkpoints = [
            V1Checkpoint.model_validate(raw_ckpt)
            for raw_ckpt in client.list_checkpoints(
                category, limit=limit, deleted=deleted
            )
        ]
        return checkpoints

    @staticmethod
    def get(id: UUID, *args, **kwargs) -> V1Checkpoint:
        """Gets a specific checkpoint.

        Args:
            id (UUID): ID of checkpoint to retrieve.

        Returns:
            V1Checkpoint: The retrieved checkpoint object.

        Examples:
            Basic usage:

            ```python
            import periflow as pf

            # Set up PeriFlow context.
            pf.init(
                api_key="YOUR_PERILFOW_API_KEY",
                project_name="my-project",
            )

            checkpoint = pf.Checkpoint.get(id="YOUR_CHECKPOINT_ID")
            ```

        """
        client = CheckpointClient()
        raw_ckpt = client.get_checkpoint(id)
        ckpt = V1Checkpoint.model_validate(raw_ckpt)
        if ckpt.forms:
            for file in ckpt.forms[0].files:
                file.path = strip_storage_path_prefix(file.path)
        return ckpt

    @staticmethod
    def delete(id: UUID) -> None:
        """Deletes a checkpoint.

        Args:
            id (UUID): ID of checkpoint to delete.

        Examples:
            Basic usage:

            ```python
            import periflow as pf

            # Set up PeriFlow context.
            pf.init(
                api_key="YOUR_PERILFOW_API_KEY",
                project_name="my-project",
            )

            checkpoint = pf.Checkpoint.delete(id="YOUR_CHECKPOINT_ID")
            ```

        """
        client = CheckpointClient()
        client.delete_checkpoint(id)

    @staticmethod
    def upload(
        name: str,
        source_path: str,
        iteration: Optional[int] = None,
        attr_file_path: Optional[str] = None,
        max_workers: int = min(32, (os.cpu_count() or 1) + 4),
    ) -> V1Checkpoint:
        """Creates a new checkpoint by uploading files in the local file system.

        Args:
            name (str): The name of checkpoint to create.
            source_path (str): Local path to the source file or directory to upload.
            iteration (Optional[int], optional): Trained step of the checkpoint. Defaults to None.
            attr_file_path (Optional[str], optional): Path to the checkpoint attribute YAML file. Defaults to None.
            max_workers (int, optional): The number of concurrency. Defaults to min(32, (os.cpu_count() or 1) + 4).

        Returns:
            V1Checkpoint: Created checkpoint object.

        Raises:
            NotFoundError: Raised when `source_path` does not exist.
            InvalidConfigError: Raised when the attribute file located at `attr_file_path` has invalid YAML format.
            InvalidAttributesError: Raised when the checkpoint attributes described in `attr_file_path` is in the invalid format.

        Examples:
            Basic usage:

            ```python
            import periflow as pf

            # Set up PeriFlow context.
            pf.init(
                api_key="YOUR_PERILFOW_API_KEY",
                project_name="my-project",
            )

            # Create a checkpoint by loading files located at "local/path/to/ckpt" directory.
            checkpoint = pf.Checkpoint.upload(
                name="my-checkpoint",
                source_path="local/path/to/ckpt",
                attr_file_path="local/path/to/attr.yaml",
            )
            ```

            :::note
            An example of attribute files is as follows.
            You have the flexibility to modify the value of each field according to your preferences,
            but it is mandatory to fill in every field.

            ```yaml
            # blenderbot
            model_type: blenderbot
            dtype: fp16
            head_size: 80
            num_heads: 32
            hidden_size: 2560
            ff_intermediate_size: 10240
            num_encoder_layers: 2
            num_decoder_layers: 24
            max_input_length: 128
            max_output_length: 128
            vocab_size: 8008
            eos_token: 2
            decoder_start_token: 1

            # bloom
            model_type: bloom
            dtype: fp16
            head_size: 128
            num_heads: 32
            num_layers: 30
            max_length: 2048
            vocab_size: 250880
            eos_token: 2

            # gpt
            model_type: gpt
            dtype: fp16
            head_size: 64
            num_heads: 25
            num_layers: 48
            max_length: 1024
            vocab_size: 50257
            eos_token: 50256

            # gpt-j
            model_type: gpt-j
            dtype: fp16
            head_size: 256
            rotary_dim: 64
            num_heads: 16
            num_layers: 28
            max_length: 2048
            vocab_size: 50400
            eos_token: 50256

            # gpt-neox
            model_type: gpt-neox
            dtype: fp16
            head_size: 128
            rotary_dim: 32
            num_heads: 40
            num_layers: 36
            max_length: 2048
            vocab_size: 50280
            eos_token: 0

            # llama
            model_type: llama
            dtype: fp16
            head_size: 128
            rotary_dim: 128
            num_heads: 32
            num_kv_heads: 32
            num_layers: 32
            ff_intermediate_size: 11008
            max_length: 2048
            vocab_size: 32000
            eos_token: 1

            # opt
            model_type: opt
            dtype: fp16
            head_size: 128
            num_heads: 32
            num_layers: 32
            max_length: 2048
            vocab_size: 50272
            eos_token: 2

            # t5
            model_type: t5
            dtype: fp16
            head_size: 128
            num_heads: 32
            hidden_size: 1024
            ff_intermediate_size: 16384
            num_encoder_layers: 24
            num_decoder_layers: 24
            max_input_length: 512
            max_output_length: 512
            num_pos_emb_buckets: 32
            max_pos_distance: 128
            vocab_size: 32100
            eos_token: 1
            decoder_start_token: 0

            # t5-v1_1
            model_type: t5-v1_1
            dtype: fp16
            head_size: 64
            num_heads: 32
            hidden_size: 2048
            ff_intermediate_size: 5120
            num_encoder_layers: 24
            num_decoder_layers: 24
            max_input_length: 512
            max_output_length: 512
            num_pos_emb_buckets: 32
            max_pos_distance: 128
            vocab_size: 32128
            eos_token: 1
            decoder_start_token: 0
            ```
            :::

        """
        expand = source_path.endswith("/") or os.path.isfile(source_path)
        src_path: Path = Path(source_path)
        if not src_path.exists():
            raise NotFoundError(f"The source path({src_path}) does not exist.")

        dist_config = {
            "pp_degree": 1,
            "dp_degree": 1,
            "mp_degree": 1,
            "dp_mode": "allreduce",
            "parallelism_order": ["pp", "dp", "mp"],
        }

        attr = {}
        if attr_file_path is not None:
            try:
                with open(attr_file_path, "r", encoding="utf-8") as attr_f:
                    attr = yaml.safe_load(attr_f)
            except yaml.YAMLError as exc:
                raise InvalidConfigError(
                    f"The attribute YAML file has invalid format: {str(exc)}"
                ) from exc

        if attr:
            validate_checkpoint_attributes(attr)

        client = CheckpointClient()
        form_client = CheckpointFormClient()
        group_client = GroupProjectCheckpointClient()
        raw_ckpt = group_client.create_checkpoint(
            name=name,
            vendor=StorageType.FAI,
            region="",
            credential_id=None,
            iteration=iteration,
            storage_name="",
            files=[],
            dist_config=dist_config,
            attributes=attr,
        )
        ckpt = V1Checkpoint.model_validate(raw_ckpt)
        if not ckpt.forms:
            raise PeriFlowInternalError(
                f"No attached model forms to the checkpoint '{ckpt.id}'"
            )
        ckpt_form_id = ckpt.forms[0].id

        try:
            logger.info("Start uploading objects to create a checkpoint(%s)...", name)
            spu_local_paths = expand_paths(src_path, FileSizeType.SMALL)
            mpu_local_paths = expand_paths(src_path, FileSizeType.LARGE)
            src_path = src_path if expand else src_path.parent
            # TODO: Need to support distributed checkpoints for model parallelism.
            spu_storage_paths = [
                attach_storage_path_prefix(
                    path=str(Path(p).relative_to(src_path)),
                    iteration=iteration or 0,
                    mp_rank=0,
                    mp_degree=1,
                    pp_rank=0,
                    pp_degree=1,
                )
                for p in spu_local_paths
            ]
            mpu_storage_paths = [
                attach_storage_path_prefix(
                    path=str(Path(p).relative_to(src_path)),
                    iteration=iteration or 0,
                    mp_rank=0,
                    mp_degree=1,
                    pp_rank=0,
                    pp_degree=1,
                )
                for p in mpu_local_paths
            ]
            spu_url_dicts = (
                form_client.get_spu_urls(
                    obj_id=ckpt_form_id, storage_paths=spu_storage_paths
                )
                if len(spu_storage_paths) > 0
                else []
            )
            mpu_url_dicts = (
                form_client.get_mpu_urls(
                    obj_id=ckpt_form_id,
                    local_paths=mpu_local_paths,
                    storage_paths=mpu_storage_paths,
                )
                if len(mpu_storage_paths) > 0
                else []
            )

            form_client.upload_files(
                obj_id=ckpt_form_id,
                spu_url_dicts=spu_url_dicts,
                mpu_url_dicts=mpu_url_dicts,
                source_path=src_path,
                max_workers=max_workers,
            )

            files = [
                get_file_info(url_info["path"], src_path) for url_info in spu_url_dicts
            ]
            files.extend(
                [
                    get_file_info(url_info["path"], src_path)
                    for url_info in mpu_url_dicts
                ]
            )
            form_client.update_checkpoint_files(ckpt_form_id=ckpt_form_id, files=files)
        except Exception as exc:
            client.delete_checkpoint(checkpoint_id=ckpt.id)
            raise exc

        logger.info(
            "Objects are uploaded and checkpoint(%s) is successfully created!", name
        )

        # Visualize the uploaded checkpoint info
        raw_ckpt = client.get_checkpoint(ckpt.id)
        ckpt = V1Checkpoint.model_validate(raw_ckpt)
        # Serving model info.
        if ckpt.forms:
            for file in ckpt.forms[0].files:
                file.path = strip_storage_path_prefix(file.path)

        return ckpt

    @staticmethod
    def download(id: UUID, save_dir: Optional[str] = None) -> None:
        """Downloads a checkpoint to the local machine.

        Args:
            id (UUID): ID of checkpoint to donwload.
            save_dir (Optional[str], optional): Local direcotry path to save the checkpoint files. Defaults to None.

        Raises:
            NotFoundError: Raised when `save_dir` is not found.

        Examples:
            Basic usage:

            ```python
            import periflow as pf

            # Set up PeriFlow context.
            pf.init(
                api_key="YOUR_PERILFOW_API_KEY",
                project_name="my-project",
            )

            pf.Checkpoint.download(
                id="YOUR_CHECKPOINT_ID",
                save_dir="local/save/dir",
            )
            ```

        """
        if save_dir is not None and not os.path.isdir(save_dir):
            raise NotFoundError(f"Directory {save_dir} is not found.")

        save_dir = save_dir or os.getcwd()

        client = CheckpointClient()
        form_client = CheckpointFormClient()
        ckpt_form_id = client.get_first_checkpoint_form(id)
        files = form_client.get_checkpoint_download_urls(ckpt_form_id)

        for i, file in enumerate(files):
            logger.info("Downloading files {%d}/{%d}...", i + 1, len(files))
            download_file(
                url=file["download_url"],
                out=os.path.join(save_dir, strip_storage_path_prefix(file["path"])),
            )

    @staticmethod
    def restore(id: UUID) -> V1Checkpoint:
        """Restores a soft-deleted checkpoint.

        Args:
            id (UUID): ID of checkpoint to restore.

        Raises:
            NotFoundError: Raised when the checkpoint is not deleted.

        Returns:
            Dict[str, Any]: The restored checkpoint info.

        Examples:
            Basics usage:

            ```python
            import periflow as pf

            # Set up PeriFlow context.
            pf.init(
                api_key="YOUR_PERILFOW_API_KEY",
                project_name="my-project",
            )

            pf.Checkpoint.download(
                id="YOUR_CHECKPOINT_ID",
                save_dir="local/save/dir",
            )
            ```

            :::info
            When a checkpoint is deleted, it becomes "soft-deleted", meaning it is
            recoverable within the 24-hour retention period. After the retention period,
            the checkpoint is "hard-deleted" and cannot be restored.
            :::

        """
        client = CheckpointClient()

        raw_ckpt = client.get_checkpoint(id)
        ckpt = V1Checkpoint.model_validate(raw_ckpt)
        if not ckpt.deleted:
            msg = f"Checkpoint({id}) is not deleted."
            raise NotFoundError(msg)

        raw_ckpt = client.restore_checkpoint(id)
        ckpt = V1Checkpoint.model_validate(raw_ckpt)
        return ckpt

    @staticmethod
    def convert(
        model_name_or_path: str,
        model_output_path: str,
        data_type: CheckpointDataType,
        *,
        tokenizer_output_dir: Optional[str] = None,
        attr_output_path: Optional[str] = None,
        cache_dir: Optional[str] = None,
        dry_run: bool = False,
    ) -> None:
        """Convert Hugging Face model checkpoint to PeriFlow format.

        Args:
            model_name_or_path (str): Hugging Face model name or local path to the checkpoint.
            model_output_path (str): Path to the converted checkpoint to save.
            data_type (CheckpointDataType): Converted checkpoint data type.
            tokenizer_output_dir (Optional[str], optional): Directory path to save 'tokenizer.json' for the converted checkpoint. Defaults to None.
            attr_output_path (Optional[str], optional): Path to create the attribute YAML file for the converted checkpoint. Defaults to None.
            cache_dir (Optional[str], optional): Path for downloading checkpoint. Defaults to None.
            dry_run (bool, optional): Check only if checkpoint is convertable. Defaults to False.

        Raises:
            NotFoundError: Raised when `model_name_or_path` is not found. Also raised when `tokenizer_output_dir` is not found.
            CheckpointConversionError: Raised when given model architecture from checkpoint is not supported to convert.

        """
        try:
            import torch  # type: ignore[import] # pylint: disable=import-outside-toplevel
        except ImportError as exc:
            raise CheckpointConversionError(
                "To convert the checkpoint, you must install 'torch'."
            ) from exc

        try:
            import transformers  # type: ignore[import] # pylint: disable=import-outside-toplevel

            from periflow.converter.maps import (  # pylint: disable=import-outside-toplevel
                model_arch_converter_map,
            )
            from periflow.converter.utils import (  # pylint: disable=import-outside-toplevel
                save_tokenizer,
            )
        except ImportError as exc:
            raise CheckpointConversionError(
                "To convert the checkpoint, your must install the package with 'pip install periflow-client[mllib]'"
            ) from exc

        data_type = validate_enums(data_type, CheckpointDataType)

        try:
            config = transformers.AutoConfig.from_pretrained(
                model_name_or_path, cache_dir=cache_dir, trust_remote_code=True
            )
        except OSError as exc:  # from transformers.AutoConfig.from_pretrained()
            config_dir = Path(model_name_or_path)
            model_output_dir = Path(model_output_path).parent
            if (
                config_dir.exists()
                and model_output_dir.absolute() == config_dir.absolute()
            ):
                raise NotFoundError(
                    f"'output_dir' ({model_output_dir.as_posix()}) and "
                    f"'model_name_or_path' ({model_name_or_path}) are the same. "
                    "In such a case, checkpoints should be prepared in 'output_dir'."
                ) from exc
            raise NotFoundError(str(exc)) from exc

        try:
            generation_config = transformers.GenerationConfig.from_pretrained(
                model_name_or_path, cache_dir=cache_dir, trust_remote_code=True
            )
        except (OSError, TypeError) as exc:
            generation_config = None

        try:
            model_arch_list = cast(
                List[str], cast(transformers.PretrainedConfig, config).architectures
            )
        except AttributeError as exc:
            raise CheckpointConversionError(str(exc)) from exc

        if len(model_arch_list) == 0:
            raise NotSupportedCheckpointError(
                invalid_option=f"'architectures={model_arch_list}'",
                valid_options=list(model_arch_converter_map.keys()),
            )
        model_arch = model_arch_list[0]
        if model_arch not in model_arch_converter_map:
            raise NotSupportedCheckpointError(
                invalid_option=f"'architectures={model_arch}'",
                valid_options=list(model_arch_converter_map.keys()),
            )

        hf_factory, converter_factory = model_arch_converter_map[model_arch]
        converter = converter_factory(
            config=config,
            generation_config=generation_config,
            output_path=model_output_path,
            data_type=data_type,
        )
        converter.check_config()

        if not dry_run:
            logger.info(
                "Start loading Hugging Face checkpoint(%s) for conversion...",
                model_name_or_path,
            )
            state_dict = hf_factory.from_pretrained(
                model_name_or_path,
                torch_dtype=torch.float32,
                cache_dir=cache_dir,
                trust_remote_code=True,
                low_cpu_mem_usage=True,  # For model loading faster and using ~1x model size CPU memory. https://huggingface.co/docs/transformers/main_classes/model#transformers.PreTrainedModel.from_pretrained.example
            ).state_dict()
            logger.info(
                "Hugging Face checkpoint(%s) is successfully loaded!",
                model_name_or_path,
            )
            converter.convert(state_dict=state_dict)
            logger.info(
                "Hugging Face checkpoint(%s) is successfully converted to Periflow format!",
                model_name_or_path,
            )

        if attr_output_path is not None:
            attr = converter.get_attributes()
            with open(attr_output_path, "w", encoding="utf-8") as file:
                yaml.dump(attr, file, sort_keys=False)

        if tokenizer_output_dir is not None:
            try:
                save_tokenizer(
                    model_name_or_path=model_name_or_path,
                    cache_dir=cache_dir,
                    save_dir=tokenizer_output_dir,
                )
            except TokenizerNotFoundError as exc:
                logger.warn(str(exc))



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""PeriFlow DeploymentClient Service."""

from __future__ import annotations

from datetime import datetime
from string import Template
from typing import Any, Callable, Dict, List, Optional

from requests import Response

from periflow.client.base import Client, ProjectRequestMixin, safe_request
from periflow.utils.url import get_serving_uri


# TODO (ym): Replace this with periflow.utils.request.paginated_get after unifying schema
def paginated_get(
    response_getter: Callable[..., Response],
    path: Optional[str] = None,
    limit: int = 20,
    **params,
) -> List[Dict[str, Any]]:
    """Pagination listing."""
    page_size = min(10, limit)
    params = {"page_size": page_size, **params}
    response_dict = response_getter(path=path, params={**params}).json()
    items = response_dict["deployments"]
    next_cursor = response_dict["cursor"]

    while next_cursor is not None and len(items) < limit:
        response_dict = response_getter(
            path=path, params={**params, "cursor": next_cursor}
        ).json()
        items.extend(response_dict["deployments"])
        next_cursor = response_dict["cursor"]

    return items


class DeploymentClient(Client[str]):
    """Deployment client."""

    @property
    def url_path(self) -> Template:
        """Get an URL path."""
        return Template(get_serving_uri("deployment/"))

    def get_deployment(self, deployment_id: str) -> Dict[str, Any]:
        """Get a deployment info."""
        response = safe_request(
            self.retrieve,
            err_prefix=f"Deployment ({deployment_id}) is not found. You may entered wrong ID.",
        )(pk=deployment_id)
        return response.json()

    def create_deployment(self, config: Dict[str, Any]) -> Dict[str, Any]:
        """Create a new deployment."""
        response = safe_request(self.post, err_prefix="Failed to post new deployment.")(
            json=config
        )
        return response.json()

    def update_deployment_scaler(
        self, deployment_id: str, min_replicas: int, max_replicas: int
    ) -> None:
        """Update a deployment auto-scaler config."""
        json_body = {
            "scaler_config": {
                "min_replica_count": min_replicas,
                "max_replica_count": max_replicas,
            },
            "update_msg": f"Set min_replicas to {min_replicas}, max_replicas to {max_replicas}",
        }
        safe_request(
            self.partial_update,
            err_prefix=f"Failed to update scaler of deployment ({deployment_id}).",
        )(pk=deployment_id, path="scaler", json=json_body)

    def list_deployments(
        self, project_id: Optional[str], archived: bool, limit: int, from_oldest: bool
    ) -> List[Dict[str, Any]]:
        """List all deployments."""
        params: Dict[str, Any] = {"archived": archived}
        if project_id:
            params["project_id"] = project_id
        if from_oldest:
            params["descending"] = False

        return paginated_get(
            safe_request(self.list, err_prefix="Failed to list deployments."),
            limit=limit,
            **params,
        )

    def stop_deployment(self, deployment_id: str) -> None:
        """Delete a deployment."""
        safe_request(self.delete, err_prefix="Failed to delete deployment.")(
            pk=deployment_id
        )


class DeploymentLogClient(Client[str]):
    """Deployment log client."""

    @property
    def url_path(self) -> Template:
        """Get an URL path."""
        return Template(get_serving_uri("deployment/$deployment_id/log/"))

    def get_deployment_logs(
        self, deployment_id: str, replica_index: int
    ) -> List[Dict[str, Any]]:
        """Get logs from a deployment."""
        response = safe_request(
            self.list,
            err_prefix=f"Log is not available for Deployment ({deployment_id})"
            f"with replica {replica_index}."
            "You may entered wrong ID or the replica is not running.",
        )(params={"replica_index": replica_index})
        return response.json()


class DeploymentMetricsClient(Client):
    """Deployment metrics client."""

    @property
    def url_path(self) -> Template:
        """Get an URL path."""
        return Template(get_serving_uri("deployment/$deployment_id/metrics/"))

    def get_metrics(self, deployment_id: str, time_window: int) -> Dict[str, Any]:
        """Get metrics from a deployment."""
        response = safe_request(
            self.list,
            err_prefix=f"Deployment ({deployment_id}) is not found. You may entered wrong ID.",
        )(data=str(time_window))
        return response.json()


class DeploymentEventClient(Client):
    """Deployment event client."""

    @property
    def url_path(self) -> Template:
        """Get an URL path."""
        return Template(get_serving_uri("deployment/$deployment_id/event/"))

    def get_events(self, deployment_id: str) -> List[Dict[str, Any]]:
        """Get deployment events."""
        response = safe_request(
            self.list,
            err_prefix=f"Events for deployment ({deployment_id}) is not found.",
        )()
        return response.json()


class DeploymentReqRespClient(Client):
    """Deployment request-response client."""

    @property
    def url_path(self) -> Template:
        """Get an URL path."""
        return Template(get_serving_uri("deployment/$deployment_id/req_resp/download/"))

    def get_download_urls(
        self, deployment_id: str, start: datetime, end: datetime
    ) -> list[dict[str, str]]:
        """Get presigned URLs to download request-response logs."""
        params = {
            "start": start.isoformat(),
            "end": end.isoformat(),
        }
        response = safe_request(
            self.list,
            err_prefix=f"Request-response logs for deployment({deployment_id}) are not found.",
        )(params=params)
        return response.json()


class PFSProjectUsageClient(Client[str], ProjectRequestMixin):
    """Project-level deployment usage client for serving."""

    def __init__(self, **kwargs):
        """Initialize project usage client for serving."""
        self.initialize_project()
        super().__init__(project_id=self.project_id, **kwargs)

    @property
    def url_path(self) -> Template:
        """Get an URL path."""
        return Template(get_serving_uri("usage/project/$project_id/duration"))

    def get_usage(
        self,
        start_date: datetime,
        end_date: datetime,
    ) -> Dict[str, Any]:
        """Get deployment usage info."""
        params = {
            "start_date": start_date.isoformat(),
            "end_date": end_date.isoformat(),
        }
        response = safe_request(
            self.list,
            err_prefix="Deployment usages are not found in the project.",
        )(params=params)
        return response.json()


class PFSVMClient(Client):
    """VM client for serving."""

    @property
    def url_path(self) -> Template:
        """Get an URL path."""
        return Template(get_serving_uri("vm/"))

    def list_vms(self) -> List[Dict[str, Any]]:
        """List all VM info."""
        response = safe_request(
            self.list,
            err_prefix="Cannot get available vm list from PFS server.",
        )()
        return response.json()



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""PeriFlow ProjectClient Service."""

# pylint: disable=too-many-arguments

from __future__ import annotations

from string import Template
from typing import Any, Dict, List, Optional
from uuid import UUID

from requests import HTTPError

from periflow.client.base import Client, ProjectRequestMixin, safe_request
from periflow.enums import CredType
from periflow.utils.format import secho_error_and_exit
from periflow.utils.maps import cred_type_map
from periflow.utils.request import paginated_get
from periflow.utils.url import get_auth_uri


def find_project_id(projects: List[Dict[str, Any]], project_name: str) -> UUID:
    """Find an ID of project among a list of projects."""
    for project in projects:
        if project["name"] == project_name:
            return UUID(project["id"])
    secho_error_and_exit(f"No project exists with name {project_name}.")


class ProjectClient(Client[UUID]):
    """Project client."""

    @property
    def url_path(self) -> Template:
        """Get an URL path."""
        return Template(get_auth_uri("pf_project"))

    def get_project(self, pf_project_id: UUID) -> Dict[str, Any]:
        """Get project info."""
        response = safe_request(self.retrieve, err_prefix="Failed to get a project.")(
            pk=pf_project_id
        )
        return response.json()

    def check_project_membership(self, pf_project_id: UUID) -> bool:
        """Check accessibility to the project."""
        try:
            self.retrieve(pf_project_id)
        except HTTPError:
            return False
        return True

    def delete_project(self, pf_project_id: UUID) -> None:
        """Delete a project."""
        safe_request(self.delete, err_prefix="Failed to delete a project.")(
            pk=pf_project_id
        )

    def list_users(self, pf_project_id: UUID) -> List[Dict[str, Any]]:
        """List all project members."""
        get_response_dict = safe_request(
            self.list, err_prefix="Failed to list users in the current project"
        )
        return paginated_get(get_response_dict, path=f"{pf_project_id}/pf_user")


class ProjectCredentialClient(Client, ProjectRequestMixin):
    """Project credential client."""

    def __init__(self, **kwargs):
        """Initialize project credential client."""
        self.initialize_project()
        super().__init__(project_id=self.project_id, **kwargs)

    @property
    def url_path(self) -> Template:
        """Get an URL path."""
        return Template(get_auth_uri("pf_project/$project_id/credential"))

    def list_credentials(
        self, cred_type: Optional[CredType] = None
    ) -> List[Dict[str, Any]]:
        """List all credentials."""
        params = {}
        if cred_type is not None:
            params["type"] = cred_type_map[cred_type]
        response = safe_request(
            self.list, err_prefix=f"Failed to list credential for {cred_type}."
        )(params=params)
        return response.json()

    def create_credential(
        self, cred_type: CredType, name: str, type_version: int, value: Dict[str, Any]
    ) -> Dict[str, Any]:
        """Create a new credential."""
        type_name = cred_type_map[cred_type]
        request_data = {
            "type": type_name,
            "name": name,
            "type_version": type_version,
            "value": value,
        }
        response = safe_request(
            self.post, err_prefix="Failed to create user credential."
        )(json=request_data)
        return response.json()



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""PeriFlow Credential Configurator."""

from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Dict, Optional, Tuple, cast
from uuid import UUID

import typer
from click import Choice
from jsonschema import Draft7Validator, ValidationError

from periflow.client.credential import CredentialClient, CredentialTypeClient
from periflow.configurator.base import InteractiveConfigurator
from periflow.enums import CredType
from periflow.utils.format import secho_error_and_exit
from periflow.utils.maps import cred_type_map_inv


@dataclass
class CredentialInteractiveConfigurator(InteractiveConfigurator[Tuple[Any, ...]]):
    """Credential configuration service."""

    ready: bool = False
    name: Optional[str] = None
    cred_type: Optional[CredType] = None
    value: Optional[Dict[str, Any]] = None

    def start_interaction(self) -> None:
        """Start interactive prompt to create a credential."""
        self.name = typer.prompt(
            "Enter the name of your new credential.", prompt_suffix="\n>> "
        )
        self.cred_type = typer.prompt(
            "What kind of credential do you want to create?\n",
            type=Choice([e.value for e in CredType]),
            prompt_suffix="\n>> ",
        )
        cred_type_client = CredentialTypeClient()
        schema = cred_type_client.get_schema_by_type(cast(CredType, self.cred_type))
        assert schema is not None
        properties: Dict[str, Any] = schema["properties"]
        self.value = {}
        typer.echo("Please fill in the following fields")
        for field, field_info in properties.items():
            field_info_str = "\n".join(f"    - {k}: {v}" for k, v in field_info.items())
            hide_input = "password" in field
            entered = typer.prompt(
                f"  {field}:\n{field_info_str}",
                prompt_suffix="\n  >> ",
                hide_input=hide_input,
            )
            self.value[field] = entered

        self._validate_schema(schema)
        self.ready = True

    def start_interaction_for_update(self, credential_id: UUID) -> None:
        """Start interactive prompt to update credential."""
        cred_client = CredentialClient()
        prev_cred = cred_client.get_credential(credential_id)

        self.name = typer.prompt(
            "Enter the NEW name of your credential. Press ENTER if you don't want to update this.\n"
            f"Current: {prev_cred['name']}",
            prompt_suffix="\n>> ",
            default=prev_cred["name"],
            show_default=False,
        )
        self.cred_type = cred_type_map_inv[prev_cred["type"]]
        cred_type_client = CredentialTypeClient()
        schema = cred_type_client.get_schema_by_type(self.cred_type)
        assert schema is not None
        properties: Dict[str, Any] = schema["properties"]
        self.value = {}
        typer.echo("Please fill in the following fields")
        for field, field_info in properties.items():
            field_info_str = "\n".join(f"    - {k}: {v}" for k, v in field_info.items())
            hide_input = "password" in field
            entered = typer.prompt(
                f"  {field} (Current: {prev_cred['value'][field]}):\n{field_info_str}",
                prompt_suffix="\n  >> ",
                default=prev_cred["value"][field],
                show_default=False,
                hide_input=hide_input,
            )
            self.value[field] = entered

        self._validate_schema(schema)
        self.ready = True

    def _validate_schema(self, schema: Dict[str, Any]) -> None:
        try:
            Draft7Validator(schema).validate(self.value)
        except ValidationError as exc:
            secho_error_and_exit(
                f"Format of credential value is invalid...! ({exc.message})"
            )

    def render(self) -> Tuple[Any, ...]:
        """Render the configured info."""
        assert self.ready
        return self.name, self.cred_type, self.value



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""PeriFlow CheckpointClient Service."""

from __future__ import annotations

from string import Template
from typing import Any, Dict, List
from uuid import UUID

from requests.models import Response

from periflow.client.base import Client, UploadableClient, safe_request
from periflow.utils.url import get_mr_uri


class CheckpointClient(Client[UUID]):
    """Checkpoint client."""

    @property
    def url_path(self) -> Template:
        """Get an URL path."""
        return Template(get_mr_uri("models/"))

    def get_checkpoint(self, checkpoint_id: UUID) -> Dict[str, Any]:
        """Get a checkpoint info."""
        response = safe_request(
            self.retrieve, err_prefix="Failed to get info of checkpoint"
        )(pk=checkpoint_id)
        return response.json()

    def get_first_checkpoint_form(self, checkpoint_id: UUID) -> UUID:
        """Get the first form of the checkpoint."""
        response = safe_request(
            self.retrieve, err_prefix="Failed to get info of checkpoint."
        )(pk=checkpoint_id)
        return UUID(response.json()["forms"][0]["id"])

    def delete_checkpoint(self, checkpoint_id: UUID) -> Response:
        """Delete a checkpoint."""
        response = safe_request(self.delete, err_prefix="Failed to delete checkpoint.")(
            pk=checkpoint_id
        )
        return response

    def restore_checkpoint(self, checkpoint_id: UUID) -> Dict[str, Any]:
        """Restore a soft-deleted checkpoint."""
        response = safe_request(self.post, err_prefix="Fail to restore checkpoint.")(
            path=f"{checkpoint_id}/restore/"
        )
        return response.json()


class CheckpointFormClient(UploadableClient[UUID]):
    """Checkpoint form client."""

    @property
    def url_path(self) -> Template:
        """Get an URL path."""
        return Template(get_mr_uri("model_forms/"))

    def update_checkpoint_files(
        self,
        ckpt_form_id: UUID,
        files: List[Dict[str, Any]],
    ) -> Dict[str, Any]:
        """Update checkpoint file metadata."""
        response = safe_request(
            self.partial_update, err_prefix="Cannot update checkpoint."
        )(pk=ckpt_form_id, json={"files": files})
        return response.json()

    def get_checkpoint_download_urls(self, ckpt_form_id: UUID) -> List[Dict[str, Any]]:
        """Get presigned URLs to download a model checkpoint."""
        response = safe_request(
            self.retrieve, err_prefix="Failed to get presigned URLs."
        )(pk=ckpt_form_id, path="download/")
        return response.json()["files"]



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""PeriFlow Credential SDK."""

# pylint: disable=line-too-long, arguments-differ, too-many-arguments, redefined-builtin

from __future__ import annotations

from typing import Any, Dict, List, Optional
from uuid import UUID

from pydantic import ValidationError

from periflow.client.credential import CredentialClient
from periflow.client.project import ProjectCredentialClient
from periflow.enums import CredType
from periflow.errors import InvalidConfigError, NotSupportedError
from periflow.schema.resource.v1.credential import V1Credential
from periflow.sdk.resource.base import ResourceAPI
from periflow.utils.maps import cred_schema_map, cred_type_map_inv


class Credential(ResourceAPI[V1Credential, UUID]):
    """PeriFlow credential resource API."""

    @staticmethod
    def create(cred_type: CredType, name: str, value: Dict[str, Any]) -> V1Credential:
        """Creates a new credential.

        Args:
            cred_type (CredType): The type of credential.
            name (str): The name of credential to create.
            value (Dict[str, Any]): Values of credential in plaintext.

        Raises:
            InvalidConfigError: Raised if `value` does not match the credential schema.

        Returns:
            V1Credential: The created credential object.

        """
        cred_schema = cred_schema_map[cred_type]
        try:
            cred_schema.model_validate(obj=value)
        except ValidationError as exc:
            raise InvalidConfigError(
                detail=f"Wrong credential format: {str(exc)}"
            ) from exc

        client = ProjectCredentialClient()
        raw_cred = client.create_credential(cred_type, name, 1, value)
        raw_cred["type"] = cred_type_map_inv[raw_cred["type"]].value
        cred = V1Credential.model_validate(raw_cred)
        return cred

    @staticmethod
    def get(id: UUID, *args, **kwargs) -> V1Credential:
        """[skip-doc] Gets a credential info.

        Args:
            id (UUID): ID of credential to retrieve.

        Returns:
            Dict[str, Any]: Retrieved credential info.

        """
        msg = "Getting a specific credential info is not supported."
        raise NotSupportedError(msg)

    @staticmethod
    def list(cred_type: Optional[CredType] = None) -> List[V1Credential]:
        """Lists credentials.

        Args:
            cred_type (Optional[CredType], optional): Filters by credential types. Defaults to None.

        Returns:
            List[V1Credential]: A list of retrieved credentials.

        """
        client = ProjectCredentialClient()
        raw_creds = client.list_credentials(cred_type)
        creds = []
        for raw_cred in raw_creds:
            cred = V1Credential.model_validate(raw_cred)
            cred.type = cred_type_map_inv[cred.type].value
            creds.append(cred)
        return creds

    @staticmethod
    def edit(
        id: UUID,
        name: Optional[str] = None,
        value: Optional[Dict[str, Any]] = None,
    ) -> V1Credential:
        """Edits credential info.

        Args:
            id (UUID): ID of credential to edit.
            name (Optional[str], optional): Name to update. No update if it is `None`. Defaults to None.
            value (Optional[Dict[str, Any]], optional): The new value of credential in plaintext. No update if it is `None`. Defaults to None.

        Returns:
            V1Credential: Updated credential object.

        """
        client = CredentialClient()
        raw_cred = client.update_credential(
            id, name=name, type_version="1", value=value
        )
        cred = V1Credential.model_validate(raw_cred)
        cred.type = cred_type_map_inv[cred.type].value
        return cred

    @staticmethod
    def delete(id: UUID) -> None:
        """Deletes a credential.

        Args:
            id (UUID): ID of credential to delete.

        """
        client = CredentialClient()
        client.delete_credential(id)



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""PeriFlow client enums."""


from __future__ import annotations

from enum import Enum, IntEnum


class GroupRole(str, Enum):
    """Organization-level roles."""

    OWNER = "owner"
    MEMBER = "member"


class ProjectRole(str, Enum):
    """Project-level roles."""

    ADMIN = "admin"
    MAINTAIN = "maintain"
    DEVELOP = "develop"
    GUEST = "guest"


class ServiceTier(str, Enum):
    """Organization service tier."""

    BASIC = "basic"
    ENT = "enterprise"


class CloudType(str, Enum):
    """Cloud types."""

    AWS = "aws"
    AZURE = "azure"
    GCP = "gcp"


class StorageType(str, Enum):
    """Cloud storage types."""

    S3 = "s3"
    BLOB = "azure-blob"
    GCS = "gcs"
    FAI = "fai"


class GpuType(str, Enum):
    """GPU types for deployment."""

    A10G = "a10g"
    A100 = "a100"
    A100_80G = "a100-80g"


class VMType(str, Enum):
    """VM types for deployment."""

    G5_XLARGE = "g5.xlarge"
    A2_HIGHGPU_1G = "a2-highgpu-1g"
    A2_ULTRAGPU_1G = "a2-ultragpu-1g"
    A2_ULTRAGPU_2G = "a2-ultragpu-2g"
    A2_ULTRAGPU_4G = "a2-ultragpu-4g"
    A2_ULTRAGPU_8G = "a2-ultragpu-8g"


class DeploymentType(str, Enum):
    """Deployment phase types."""

    DEV = "dev"
    PROD = "prod"


class DeploymentSecurityLevel(str, Enum):
    """Deployment access levels."""

    PUBLIC = "public"
    PROTECTED = "protected"


class CheckpointCategory(str, Enum):
    """Checkpoint categories."""

    USER_PROVIDED = "USER"
    JOB_GENERATED = "JOB"
    COPIED = "COPY"
    REFERENCED = "REF"


class CheckpointStatus(str, Enum):
    """Checkpoint statuses."""

    CREATED = "Created"  # Model is created
    COPYING = "Copying"  # Copying the model from catalog
    ACTIVE = "Active"  # Available to use
    FAILED = "Failed"  # Cannot use the checkpoint


class CheckpointValidationStatus(str, Enum):
    """Checkpoint validation statuses."""

    VALIDATING = "Validating"  # Validating the checkpoint format
    VALID = "Valid"  # Checkpoint format is valid
    INVALID = "Invalid"  # Checkpoint is invalid
    FAILED = "Failed"  # Validation failed for some reason


class CatalogStatus(str, Enum):
    """Catalog statuses."""

    CREATED = "Created"  # Catalog is created
    PUBLISHING = "Publishing"  # Publishing the catalog
    ACTIVE = "Active"  # Available to use the catalog
    FAILED = "Failed"  # Failed to publish catalog


class CredType(str, Enum):
    """Credential types."""

    S3 = "s3"
    BLOB = "azure-blob"
    GCS = "gcs"


class BeamSearchType(IntEnum):
    """Beam search types."""

    DETERMINISTIC = 0
    STOCHASTIC = 1
    NAIVE_SAMPLING = 2


class DeploymentReplicaStatus(str, Enum):
    """Deployment replica status types."""

    INITIALIZING = "Initializing"
    PENDING = "Pending"
    RUNNING = "Running"
    SUCCESS = "Success"
    FAILED = "Failed"
    UNKNOWN = "Unknown"
    STOPPING = "Stopping"
    TERMINATED = "Terminated"


class DeploymentStatus(str, Enum):
    """Deployment status types."""

    INITIALIZING = "Initializing"
    HEALTHY = "Healthy"
    UNHEALTHY = "Unhealthy"
    STOPPING = "Stopping"
    TERMINATED = "Terminated"


class CheckpointDataType(str, Enum):
    """Checkpoint dtype enums."""

    BF16 = "bf16"
    FP16 = "fp16"
    FP32 = "fp32"



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""PeriFlow CLI API Request Utilities."""

from __future__ import annotations

from typing import Any, Callable, Dict, List, Optional

from requests.exceptions import HTTPError
from requests.models import Response

from periflow.utils.url import discuss_url

DEFAULT_PAGINATION_SIZE = 50
DEFAULT_REQ_TIMEOUT = 30


def decode_http_err(exc: HTTPError) -> str:
    """Decode HTTP error."""
    try:
        if exc.response.status_code == 500:
            error_str = f"Internal Server Error: Please contact to system admin via {discuss_url}"
        elif exc.response.status_code == 404:
            error_str = (
                "Not Found: The requested resource is not found. Please check it again. "
                f"If you cannot find out why this error occurs, please visit {discuss_url}."
            )
        else:
            response = exc.response
            detail_json = response.json()
            if "detail" in detail_json:
                error_str = f"Error Code: {response.status_code}\nDetail: {detail_json['detail']}"
            elif "error_description" in detail_json:
                error_str = (
                    f"Error Code: {response.status_code}\n"
                    f"Detail: {detail_json['error_description']}"
                )
            else:
                error_str = f"Error Code: {response.status_code}"
    except ValueError:
        error_str = exc.response.content.decode()

    return error_str


def paginated_get(
    response_getter: Callable[..., Response],
    path: Optional[str] = None,
    limit: int = 20,
    **params,
) -> List[Dict[str, Any]]:
    """List objects with pagination."""
    page_size = min(DEFAULT_PAGINATION_SIZE, limit)
    params = {"limit": page_size, **params}
    response_dict = response_getter(path=path, params={**params}).json()
    items = response_dict["results"]
    next_cursor = response_dict["next_cursor"]

    while next_cursor is not None and len(items) < limit:
        response_dict = response_getter(
            path=path, params={**params, "cursor": next_cursor}
        ).json()
        items.extend(response_dict["results"])
        next_cursor = response_dict["next_cursor"]

    return items



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""PeriFlow File Service."""

from __future__ import annotations

from string import Template
from typing import Any, Dict
from uuid import UUID

from periflow.client.base import (
    Client,
    GroupRequestMixin,
    ProjectRequestMixin,
    UserRequestMixin,
    safe_request,
)
from periflow.utils.url import get_mr_uri


class FileClient(Client[UUID]):
    """File client service."""

    @property
    def url_path(self) -> Template:
        """Get an URL path."""
        return Template(get_mr_uri("files/"))

    def get_misc_file_upload_url(self, misc_file_id: UUID) -> str:
        """Get an URL to upload file.

        Args:
            misc_file_id (UUID): Misc file ID to upload.

        Returns:
            str: An uploadable URL.

        """
        response = safe_request(self.post, err_prefix="Failed to get file upload URL.")(
            path=f"{misc_file_id}/upload/"
        )
        return response.json()["upload_url"]

    def get_misc_file_download_url(self, misc_file_id: UUID) -> str:
        """Get an URL to download file.

        Args:
            misc_file_id (UUID): Misc file ID to download.

        Returns:
            Dict[str, Any]: A downloadable URL.

        """
        response = safe_request(
            self.post, err_prefix="Failed to get file download URL."
        )(path=f"{misc_file_id}/download/")
        return response.json()["download_url"]

    def make_misc_file_uploaded(self, misc_file_id: UUID) -> Dict[str, Any]:
        """Request to mark the file as uploaded.

        Args:
            misc_file_id (UUID): Misc file ID to change status.

        Returns:
            Dict[str, Any]: The updated file info.

        """
        response = safe_request(
            self.partial_update, err_prefix="Failed to patch the file status."
        )(pk=misc_file_id, path="uploaded/")
        return response.json()


class GroupProjectFileClient(
    Client, UserRequestMixin, GroupRequestMixin, ProjectRequestMixin
):
    """Group-shared file client."""

    def __init__(self, **kwargs):
        """Initialize organization project file client."""
        self.initialize_user()
        self.initialize_group()
        self.initialize_project()
        super().__init__(group_id=self.group_id, project_id=self.project_id, **kwargs)

    @property
    def url_path(self) -> Template:
        """Get an URL path."""
        return Template(get_mr_uri("orgs/$group_id/prjs/$project_id/files/"))

    def create_misc_file(self, file_info: Dict[str, Any]) -> Dict[str, Any]:
        """Request to create a misc file.

        Args:
            file_info (Dict[str, Any]): File info.

        Returns:
            Dict[str, Any]: Response body with the created file info.

        """
        request_data = {
            "user_id": str(self.user_id),
            **file_info,
        }
        response = safe_request(self.post, err_prefix="Failed to create file.")(
            json=request_data
        )
        return response.json()



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""PeriFlow Auth Tools."""

from __future__ import annotations

import functools
from enum import Enum
from typing import Any, Callable, Dict, Optional, Union

import requests

import periflow
from periflow.errors import AuthTokenNotFoundError
from periflow.utils.format import secho_error_and_exit
from periflow.utils.fs import get_periflow_directory
from periflow.utils.request import DEFAULT_REQ_TIMEOUT
from periflow.utils.url import get_training_uri

access_token_path = get_periflow_directory() / "access_token"
refresh_token_path = get_periflow_directory() / "refresh_token"
mfa_token_path = get_periflow_directory() / "mfa_token"


class TokenType(str, Enum):
    """Token type enums."""

    ACCESS = "ACCESS"
    REFRESH = "REFRESH"
    MFA = "MFA"


token_path_map = {
    TokenType.ACCESS: access_token_path,
    TokenType.REFRESH: refresh_token_path,
    TokenType.MFA: mfa_token_path,
}


def get_auth_header() -> Dict[str, Any]:
    """Get authorization header.

    Returns:
        Dict[str, Any]: HTTP Authorization headers for the request.

    """
    token: Optional[str]
    if periflow.api_key:
        token = periflow.api_key
    else:
        token = get_token(TokenType.ACCESS)

    if token is None:
        raise AuthTokenNotFoundError(
            "Should set PERIFLOW_API_KEY environment variable or sign in with 'pf login'."
        )

    return {"Authorization": f"Bearer {token}"}


def get_token(token_type: TokenType) -> Union[str, None]:
    """Get an auth token."""
    try:
        if token_type == TokenType.ACCESS:
            return access_token_path.read_text()
        if token_type == TokenType.REFRESH:
            return refresh_token_path.read_text()
        if token_type == TokenType.MFA:
            return mfa_token_path.read_text()
        secho_error_and_exit(
            "token_type should be one of 'access' or 'refresh' or 'mfa'."
        )
    except FileNotFoundError:
        return None


def update_token(token_type: TokenType, token: str) -> None:
    """Update saved token value."""
    token_path_map[token_type].write_text(token)


def delete_token(token_type: TokenType) -> None:
    """Delete token."""
    token_path_map[token_type].unlink(missing_ok=True)


def clear_tokens() -> None:
    """Clear all tokens."""
    for token_type in TokenType:
        delete_token(token_type)


# pylint: disable=too-many-nested-blocks
def auto_token_refresh(
    func: Callable[..., requests.Response]
) -> Callable[..., requests.Response]:
    """Decorator for automatic token refresh."""

    @functools.wraps(func)
    def inner(*args, **kwargs) -> requests.Response:
        resp = func(*args, **kwargs)
        if resp.status_code in (401, 403):
            refresh_token = get_token(TokenType.REFRESH)
            if refresh_token is not None:
                refresh_r = requests.post(
                    get_training_uri("token/refresh/"),
                    data={"refresh_token": refresh_token},
                    timeout=DEFAULT_REQ_TIMEOUT,
                )
                try:
                    refresh_r.raise_for_status()
                except requests.HTTPError:
                    secho_error_and_exit(
                        "Failed to refresh access token... Please login again"
                    )

                update_token(
                    token_type=TokenType.ACCESS, token=refresh_r.json()["access_token"]
                )
                update_token(
                    token_type=TokenType.REFRESH,
                    token=refresh_r.json()["refresh_token"],
                )
                # We need to restore file offset if we want to transfer file objects
                if "files" in kwargs:
                    files = kwargs["files"]
                    for _, file_tuple in files.items():
                        for element in file_tuple:
                            if hasattr(element, "seek"):
                                # Restore file offset
                                element.seek(0)
                resp = func(*args, **kwargs)
                resp.raise_for_status()
            else:
                secho_error_and_exit(
                    "Failed to refresh access token... Please login again"
                )
        else:
            resp.raise_for_status()
        return resp

    return inner



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""PeriFlow Serving API Interface."""

from __future__ import annotations

import os
from abc import ABC, abstractmethod
from contextlib import asynccontextmanager
from typing import (
    Any,
    AsyncIterator,
    Dict,
    Generic,
    Literal,
    Optional,
    TypeVar,
    Union,
    overload,
)
from urllib.parse import urljoin

import aiohttp
import requests
from pydantic import BaseModel

from periflow.auth import get_auth_header
from periflow.client.deployment import DeploymentClient
from periflow.enums import DeploymentSecurityLevel
from periflow.errors import InvalidConfigError, NotFoundError
from periflow.logging import logger
from periflow.utils.request import DEFAULT_REQ_TIMEOUT
from periflow.utils.url import get_baseurl

_Generation = TypeVar("_Generation", bound=BaseModel)
_GenerationLine = TypeVar("_GenerationLine", bound=BaseModel)
_Options = TypeVar("_Options")


class GenerationStream(ABC, Generic[_GenerationLine, _Generation]):
    """Generation stream."""

    def __init__(self, response: requests.Response) -> None:
        """Initializes generation stream."""
        self._iter = response.iter_lines()

    def __iter__(self) -> GenerationStream:  # noqa: D105
        return self

    @abstractmethod
    def __next__(self) -> _GenerationLine:
        """Iterates the stream."""

    @abstractmethod
    def wait(self) -> Optional[_Generation]:
        """Waits for the generation to complete."""


class AsyncGenerationStream(ABC, Generic[_GenerationLine, _Generation]):
    """Asynchronous generation stream."""

    def __init__(self, response: aiohttp.ClientResponse) -> None:
        """Initializes generation stream."""
        self._iter = response.content.__aiter__()

    def __aiter__(self) -> AsyncGenerationStream:  # noqa: D105
        return self

    @abstractmethod
    async def __anext__(self) -> _GenerationLine:
        """Iterates the stream."""

    @abstractmethod
    async def wait(self) -> Optional[_Generation]:
        """Waits for the generation to complete."""


_GenerationStream = TypeVar("_GenerationStream", bound=GenerationStream)
_AsyncGenerationStream = TypeVar("_AsyncGenerationStream", bound=AsyncGenerationStream)


class ServingAPI(
    ABC, Generic[_Generation, _GenerationStream, _AsyncGenerationStream, _Options]
):
    """Serving API interface."""

    def __init__(
        self,
        deployment_id: Optional[str] = None,
        endpoint: Optional[str] = None,
        deployment_security_level: Optional[DeploymentSecurityLevel] = None,
    ) -> None:
        """Initializes ServingAPI."""
        if deployment_id is None and endpoint is None:
            raise InvalidConfigError(
                "One of 'deployment_id' and 'endpoint' should be provided."
            )
        if deployment_id is not None and endpoint is not None:
            raise InvalidConfigError(
                "Only provide one between 'deployment_id' and 'endpoint'."
            )

        if deployment_id is not None:
            client = DeploymentClient()
            deployment = client.get_deployment(deployment_id)
            endpoint = deployment["endpoint"]
            if not endpoint:
                raise NotFoundError("Active endpoint for the deployment is not found.")
            self._endpoint = urljoin(
                get_baseurl(endpoint), os.path.join(deployment_id, self._api_path)
            )
            self._auth_required = deployment["config"]["infrequest_perm_check"]
        elif endpoint is not None:
            if deployment_security_level is None:
                raise InvalidConfigError(
                    "'deployment_security_level' should be provided."
                )
            self._endpoint = endpoint
            self._auth_required = (
                deployment_security_level == DeploymentSecurityLevel.PROTECTED
            )

        self._session: Optional[aiohttp.ClientSession] = None

    def _get_headers(self) -> Dict[str, Any]:
        if self._auth_required:
            return get_auth_header()
        return {}

    @property
    @abstractmethod
    def _api_path(self) -> str:
        """Serving API URL path."""

    @overload
    @abstractmethod
    def create(self, options: _Options, *, stream: Literal[True]) -> _GenerationStream:
        ...

    @overload
    @abstractmethod
    def create(self, options: _Options, *, stream: Literal[False]) -> _Generation:
        ...

    @abstractmethod
    def create(
        self, options: _Options, *, stream: bool = False
    ) -> Union[_GenerationStream, _Generation]:
        """Creates a new serving result."""

    @overload
    @abstractmethod
    async def acreate(
        self, options: _Options, *, stream: Literal[True]
    ) -> _AsyncGenerationStream:
        ...

    @overload
    @abstractmethod
    async def acreate(
        self, options: _Options, *, stream: Literal[False]
    ) -> _Generation:
        ...

    @abstractmethod
    async def acreate(
        self, options: _Options, *, stream: bool = False
    ) -> Union[_AsyncGenerationStream, _Generation]:
        """Async API to create a new serving result."""

    @asynccontextmanager
    async def api_session(self) -> AsyncIterator[None]:
        """Creates a new API session."""
        if self._session is not None:
            logger.warn("API session is already opened.")
            return

        timeout = aiohttp.ClientTimeout(total=DEFAULT_REQ_TIMEOUT)
        async with aiohttp.ClientSession(
            headers=self._get_headers(), timeout=timeout
        ) as session:
            self._session = session
            try:
                yield
            finally:
                self._session = None
